/*
btreorg.c Version 0.2.0  Btree reorganize database
Copyright (C) 2003-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

int btreorg(bthdr)
btreefmt *bthdr;
   {
   int rslt;
   int lockrslt;
   int position;
   int opnerr;
   unsigned char *p,*q;
   unsigned char *rndSeed;
   btreefmt *tmpbt;
   unsigned char tmpdbname[32];
   unsigned char tmpbtdbn[32];
   unsigned char buf[4096 * 16];
   bthdr->reason = 0;
   lockrslt = lockf(bthdr->hndl,F_TEST,0);
   if (lockrslt < 0)
      {
      return(UPDTLOCK3);
      } /* can't reorg when file is locked */
   lockrslt = lockf(bthdr->hndl,F_TLOCK,0);
   if (lockrslt < 0)
      {
      fprintf(stderr,"btreorg: lock error on %s\n",
	 bthdr->dbname);
      perror("Lock error");
      return(UPDTLOCK4);
      } /* can't reorg if lock error */
   tmpbt = (btreefmt *) btopent(bthdr->nodesz,
      bthdr->rcdlen,
      bthdr->keylen,
      bthdr->keyofst,
      bthdr->dletofst,
      &opnerr,NULL);
   if (tmpbt == NULL)
      {
      fprintf(stderr,"btreorg: error initializing "
	 "the temporary database: %d\n",
	 opnerr);
      bthdr->reason = BTREE_ERR + 2601;
      return(BTREE_ERR + 2601);
      } /* can't create temp bt */
   position = SWPFRST;
   rslt = 0;
   while (!rslt)
      {
      rslt = btsweep(bthdr,position,buf);
      if (!rslt)
	 {
	 rslt = btisrt(tmpbt,ROOT,buf);
	 if (rslt)
	    {
            fprintf(stderr,"btreorg: error inserting in "
	       "the temporary database: %d reason %d\n",
	       rslt, bthdr->reason);
	    fprintf(stderr,"btreorg: "
	       "ofst %d rcdnum %d\n",
	       (unsigned int) bthdr->ofst,
	       bthdr->rcdnum);
	    bthxmem(bthdr->blksz,
               (unsigned char *) bthdr->blk);
	    btcloset(tmpbt);
            bthdr->reason = BTREE_ERR + 2602;
            return(BTREE_ERR + 2602);
	    } /* if insert err */
	 } /* if good read */
      position = SWPNEXT;
      } /* read loop */
   if (bthdr->reason)
      {
      fprintf(stderr,"btreorg: btsweep error %d\n",
	 bthdr->reason);
      return(BTREE_ERR + 2603);
      } /* btsweep err */
   rndSeed = (unsigned char *) rndinit();
   p = (unsigned char *) tmpdbname;
   q = (unsigned char *) p + 8;
   while (p < q)
      {
      *p++ = rndnum(26,rndSeed) + 'A';
      } /* for each char in temp dbname */
   *p = '\0';
   strcat((char *) tmpdbname,".BT");
   strcpy((char *) tmpbtdbn,(char *) tmpbt->dbname);
   btclose(tmpbt);
   /* rename the database name */
   /* to a temporary name */
   rslt = rename((char *) bthdr->dbname,(char *) tmpdbname);
   if (rslt < 0)
      {
      fprintf(stderr,"btreorg: error renaming %s to %s\n",
	 bthdr->dbname, tmpdbname);
      perror("Rename error");
      return(SYS_ERR + errno);
      } /* rename failed */
   /* rename the reorganized database */
   /* to the original database name */
   rslt = rename((char *) tmpbtdbn,(char *) bthdr->dbname);
   if (rslt < 0)
      {
      fprintf(stderr,"btreorg: error renaming %s to %s\n",
	 tmpbtdbn, bthdr->dbname);
      perror("Rename error");
      return(SYS_ERR + errno);
      } /* rename failed */
   /* remove the temporary name */
   /* to which the original database */
   /* was renamed */
   rslt = remove((char *) tmpdbname);
   if (rslt < 0)
      {
      fprintf(stderr,"btreorg: error removing %s\n",
	 tmpdbname);
      perror("Remove error");
      return(SYS_ERR + errno);
      } /* remove failed */
   if (bthdr->lock != LOCK)
      {
      lockrslt = lockf(bthdr->hndl,F_ULOCK,0);
      if (lockrslt < 0)
         {
         fprintf(stderr,"btreorg: error unlocking "
            "%s after reorganization\n",
            bthdr->dbname);
         perror("Unlock error");
         bthdr->reason  = SYS_ERR + errno;
	 return(SYS_ERR + errno);
         }
      } /* if no lock during open, unlock */
   return(SWPFOUND);
   } /* btreorg */
